r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.llm._base import BaseLlmAdaptor
from memori.llm._registry import Registry
from memori.llm._utils import agno_is_anthropic, llm_is_anthropic


@Registry.register_adapter(llm_is_anthropic)
@Registry.register_adapter(agno_is_anthropic)
class Adapter(BaseLlmAdaptor):
    def get_formatted_query(self, payload):
        """
        [
            {
                "content": "...",
                "role": "..."
            }
        ]
        """

        try:
            messages = payload["conversation"]["query"].get("messages", [])
            return self._exclude_injected_messages(messages, payload)
        except KeyError:
            return []

    def get_formatted_response(self, payload):
        try:
            content = payload["conversation"]["response"].get("content", None)
        except KeyError:
            return []

        response = []
        if content is not None:
            # Unstreamed
            # [
            #   {
            #       "text": "...",
            #       "type": "..."
            #   }
            # ]
            for entry in content:
                response.append(
                    {
                        "role": payload["conversation"]["response"].get("role", None),
                        "text": entry["text"],
                        "type": entry["type"],
                    }
                )
        else:
            # Streamed
            """
            REQUEST FOR CONTRIBUTION

            How do we parse the messages when Anthropic streams the response?
            """
            raise RuntimeError("REQUEST FOR CONTRIBUTION")

        return response



r"""
 _  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                 perfectam memoriam
                      memorilabs.ai
"""


class QuotaExceededError(Exception):
    def __init__(
        self,
        message=(
            "your IP address is over quota; register for an API key now: "
            + "https://app.memorilabs.ai/signup"
        ),
    ):
        self.message = message
        super().__init__(self.message)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import asyncio

from memori.llm._base import BaseClient
from memori.llm._constants import (
    AGNO_FRAMEWORK_PROVIDER,
    AGNO_GOOGLE_LLM_PROVIDER,
    ATHROPIC_LLM_PROVIDER,
    GOOGLE_LLM_PROVIDER,
    LANGCHAIN_CHATBEDROCK_LLM_PROVIDER,
    LANGCHAIN_CHATGOOGLEGENAI_LLM_PROVIDER,
    LANGCHAIN_CHATVERTEXAI_LLM_PROVIDER,
    LANGCHAIN_FRAMEWORK_PROVIDER,
    LANGCHAIN_OPENAI_LLM_PROVIDER,
    OPENAI_LLM_PROVIDER,
    PYDANTIC_AI_FRAMEWORK_PROVIDER,
    PYDANTIC_AI_OPENAI_LLM_PROVIDER,
)
from memori.llm._invoke import (
    Invoke,
    InvokeAsync,
    InvokeAsyncIterator,
    InvokeAsyncStream,
    InvokeStream,
)
from memori.llm._registry import Registry


@Registry.register_client(lambda client: hasattr(client, "messages"))
class Anthropic(BaseClient):
    def register(self, client, _provider=None):
        if not hasattr(client, "messages"):
            raise RuntimeError("client provided is not instance of Anthropic")

        if not hasattr(client, "_memori_installed"):
            client.beta._messages_create = client.beta.messages.create
            client._messages_create = client.messages.create

            try:
                asyncio.get_running_loop()

                client.beta.messages.create = (
                    InvokeAsync(self.config, client.beta._messages_create)
                    .set_client(_provider, ATHROPIC_LLM_PROVIDER, client._version)
                    .invoke
                )
                client.messages.create = (
                    InvokeAsync(self.config, client._messages_create)
                    .set_client(_provider, ATHROPIC_LLM_PROVIDER, client._version)
                    .invoke
                )
            except RuntimeError:
                client.beta.messages.create = (
                    Invoke(self.config, client.beta._messages_create)
                    .set_client(_provider, ATHROPIC_LLM_PROVIDER, client._version)
                    .invoke
                )
                client.messages.create = (
                    Invoke(self.config, client._messages_create)
                    .set_client(_provider, ATHROPIC_LLM_PROVIDER, client._version)
                    .invoke
                )

            client._memori_installed = True

        return self


@Registry.register_client(lambda client: hasattr(client, "models"))
class Google(BaseClient):
    def register(self, client, _provider=None):
        if not hasattr(client, "models"):
            raise RuntimeError("client provided is not instance of genai.Client")

        if not hasattr(client, "_memori_installed"):
            client.models.actual_generate_content = client.models.generate_content
            client_version = getattr(client, "_version", None)

            llm_provider = (
                AGNO_GOOGLE_LLM_PROVIDER
                if _provider == AGNO_FRAMEWORK_PROVIDER
                else GOOGLE_LLM_PROVIDER
            )

            client.models.generate_content = (
                Invoke(self.config, client.models.actual_generate_content)
                .set_client(_provider, llm_provider, client_version)
                .uses_protobuf()
                .invoke
            )

            # Register sync streaming if available
            if hasattr(client.models, "generate_content_stream"):
                client.models.actual_generate_content_stream = (
                    client.models.generate_content_stream
                )
                client.models.generate_content_stream = (
                    Invoke(
                        self.config,
                        client.models.actual_generate_content_stream,
                    )
                    .set_client(_provider, llm_provider, client_version)
                    .uses_protobuf()
                    .invoke
                )

            # Register async client if available
            if hasattr(client, "aio") and hasattr(client.aio, "models"):
                client.aio.models.actual_generate_content = (
                    client.aio.models.generate_content
                )
                client.aio.models.generate_content = (
                    InvokeAsync(self.config, client.aio.models.actual_generate_content)
                    .set_client(_provider, llm_provider, client_version)
                    .uses_protobuf()
                    .invoke
                )

                # Register streaming if available
                if hasattr(client.aio.models, "generate_content_stream"):
                    client.aio.models.actual_generate_content_stream = (
                        client.aio.models.generate_content_stream
                    )
                    client.aio.models.generate_content_stream = (
                        InvokeAsyncIterator(
                            self.config,
                            client.aio.models.actual_generate_content_stream,
                        )
                        .set_client(_provider, llm_provider, client_version)
                        .uses_protobuf()
                        .invoke
                    )

            client._memori_installed = True

        return self


class LangChain(BaseClient):
    def register(
        self, chatbedrock=None, chatgooglegenai=None, chatopenai=None, chatvertexai=None
    ):
        if (
            chatbedrock is None
            and chatgooglegenai is None
            and chatopenai is None
            and chatvertexai is None
        ):
            raise RuntimeError("LangChain::register called without client")

        if chatbedrock is not None:
            if not hasattr(chatbedrock, "client"):
                raise RuntimeError("client provided is not instance of ChatBedrock")

            if not hasattr(chatbedrock.client, "_memori_installed"):
                chatbedrock.client._invoke_model = chatbedrock.client.invoke_model
                chatbedrock.client.invoke_model = (
                    Invoke(self.config, chatbedrock.client._invoke_model)
                    .set_client(
                        LANGCHAIN_FRAMEWORK_PROVIDER,
                        LANGCHAIN_CHATBEDROCK_LLM_PROVIDER,
                        None,
                    )
                    .invoke
                )

                chatbedrock.client._invoke_model_with_response_stream = (
                    chatbedrock.client.invoke_model_with_response_stream
                )
                chatbedrock.client.invoke_model_with_response_stream = (
                    Invoke(
                        self.config,
                        chatbedrock.client._invoke_model_with_response_stream,
                    )
                    .set_client(
                        LANGCHAIN_FRAMEWORK_PROVIDER,
                        LANGCHAIN_CHATBEDROCK_LLM_PROVIDER,
                        None,
                    )
                    .invoke
                )

                chatbedrock.client._memori_installed = True

        if chatgooglegenai is not None:
            if not hasattr(chatgooglegenai, "client"):
                raise RuntimeError(
                    "client provided is not instance of ChatGoogleGenerativeAI"
                )

            if not hasattr(chatgooglegenai.client, "_memori_installed"):
                chatgooglegenai.client._generate_content = (
                    chatgooglegenai.client.generate_content
                )
                chatgooglegenai.client.generate_content = (
                    Invoke(self.config, chatgooglegenai.client._generate_content)
                    .set_client(
                        LANGCHAIN_FRAMEWORK_PROVIDER,
                        LANGCHAIN_CHATGOOGLEGENAI_LLM_PROVIDER,
                        None,
                    )
                    .uses_protobuf()
                    .invoke
                )

                if chatgooglegenai.async_client is not None:
                    chatgooglegenai.async_client._stream_generate_content = (
                        chatgooglegenai.async_client.stream_generate_content
                    )
                    chatgooglegenai.async_client.stream_generate_content = (
                        InvokeAsyncIterator(
                            self.config,
                            chatgooglegenai.async_client._stream_generate_content,
                        )
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_CHATGOOGLEGENAI_LLM_PROVIDER,
                            None,
                        )
                        .uses_protobuf()
                        .invoke
                    )

                chatgooglegenai.client._memori_installed = True

        if chatopenai is not None:
            if not hasattr(chatopenai, "client") or not hasattr(
                chatopenai, "async_client"
            ):
                raise RuntimeError("client provided is not instance of ChatOpenAI")

            for client in filter(
                None,
                [getattr(chatopenai, "http_client", None), chatopenai.client._client],
            ):
                if not hasattr(client, "_memori_installed"):
                    client.beta._chat_completions_create = (
                        client.beta.chat.completions.create
                    )
                    client.beta.chat.completions.create = (
                        Invoke(self.config, client.beta._chat_completions_create)
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client.beta._chat_completions_parse = (
                        client.beta.chat.completions.parse
                    )
                    client.beta.chat.completions.parse = (
                        Invoke(self.config, client.beta._chat_completions_parse)
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._chat_completions_create = client.chat.completions.create
                    client.chat.completions.create = (
                        Invoke(self.config, client._chat_completions_create)
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._chat_completions_parse = client.chat.completions.parse
                    client.chat.completions.parse = (
                        Invoke(self.config, client._chat_completions_parse)
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._memori_installed = True

            for client in filter(
                None,
                [
                    getattr(chatopenai, "async_http_client", None),
                    chatopenai.async_client._client,
                ],
            ):
                if not hasattr(client, "_memori_installed"):
                    client.beta._chat_completions_create = (
                        client.beta.chat.completions.create
                    )
                    client.beta.chat.completions.create = (
                        InvokeAsyncIterator(
                            self.config, client.beta._chat_completions_create
                        )
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client.beta._chat_completions_parse = (
                        client.beta.chat.completions.parse
                    )
                    client.beta.chat.completions.parse = (
                        InvokeAsyncIterator(
                            self.config, client.beta._chat_completions_parse
                        )
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._chat_completions_create = client.chat.completions.create
                    client.chat.completions.create = (
                        InvokeAsyncIterator(
                            self.config, client._chat_completions_create
                        )
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._chat_completions_parse = client.chat.completions.parse
                    client.chat.completions.parse = (
                        InvokeAsyncIterator(self.config, client._chat_completions_parse)
                        .set_client(
                            LANGCHAIN_FRAMEWORK_PROVIDER,
                            LANGCHAIN_OPENAI_LLM_PROVIDER,
                            None,
                        )
                        .invoke
                    )

                    client._memori_installed = True

        if chatvertexai is not None:
            if not hasattr(chatvertexai, "prediction_client"):
                raise RuntimeError("client provided isnot instance of ChatVertexAI")

            if not hasattr(chatvertexai.prediction_client, "_memori_installed"):
                chatvertexai.prediction_client.actual_generate_content = (
                    chatvertexai.prediction_client.generate_content
                )
                chatvertexai.prediction_client.generate_content = (
                    Invoke(
                        self.config,
                        chatvertexai.prediction_client.actual_generate_content,
                    )
                    .set_client(
                        LANGCHAIN_FRAMEWORK_PROVIDER,
                        LANGCHAIN_CHATVERTEXAI_LLM_PROVIDER,
                        None,
                    )
                    .uses_protobuf()
                    .invoke
                )

                chatvertexai.prediction_client._memori_installed = True

        if self.config.llm.provider is None:
            raise RuntimeError("Unsupported LLM client")

        return self


@Registry.register_client(
    lambda client: hasattr(client, "chat") and hasattr(client, "_version")
)
class OpenAi(BaseClient):
    def register(self, client, _provider=None, stream=False):
        if not hasattr(client, "chat"):
            raise RuntimeError("client provided is not instance of OpenAI")

        if not hasattr(client, "_memori_installed"):
            client.beta._chat_completions_parse = client.beta.chat.completions.parse
            client.chat._completions_create = client.chat.completions.create

            try:
                asyncio.get_running_loop()

                if stream is True:
                    client.beta.chat.completions.parse = (
                        InvokeAsyncStream(
                            self.config, client.beta._chat_completions_parse
                        )
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                    client.chat.completions.create = (
                        InvokeAsyncStream(
                            self.config,
                            client.chat._completions_create,
                        )
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                else:
                    client.beta.chat.completions.parse = (
                        InvokeAsync(self.config, client.beta._chat_completions_parse)
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                    client.chat.completions.create = (
                        InvokeAsync(
                            self.config,
                            client.chat._completions_create,
                        )
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
            except RuntimeError:
                if stream is True:
                    client.beta.chat.completions.parse = (
                        InvokeStream(self.config, client.beta._chat_completions_parse)
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                    client.chat.completions.create = (
                        InvokeStream(
                            self.config,
                            client.chat._completions_create,
                        )
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                else:
                    client.beta.chat.completions.parse = (
                        Invoke(self.config, client.beta._chat_completions_parse)
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )
                    client.chat.completions.create = (
                        Invoke(
                            self.config,
                            client.chat._completions_create,
                        )
                        .set_client(_provider, OPENAI_LLM_PROVIDER, client._version)
                        .invoke
                    )

            client._memori_installed = True

        return self


@Registry.register_client(
    lambda client: hasattr(client, "chat")
    and hasattr(client.chat, "completions")
    and not hasattr(client, "_version")
)
class PydanticAi(BaseClient):
    def register(self, client):
        if not hasattr(client, "chat"):
            raise RuntimeError("client provided was not instantiated using PydanticAi")

        if not hasattr(client, "_memori_installed"):
            client.chat.completions.actual_chat_completions_create = (
                client.chat.completions.create
            )

            client.chat.completions.create = (
                InvokeAsyncIterator(
                    self.config,
                    client.chat.completions.actual_chat_completions_create,
                )
                .set_client(
                    PYDANTIC_AI_FRAMEWORK_PROVIDER,
                    PYDANTIC_AI_OPENAI_LLM_PROVIDER,
                    client._version,
                )
                .invoke
            )

            client._memori_installed = True

        return self


@Registry.register_client(lambda client: "xai" in str(type(client).__module__).lower())
class XAi(BaseClient):
    """
    XAI client requires special handling due to its two-step API.

    Unlike other clients, the actual API call happens on the Chat object
    returned by create(), not on the create() method itself. All wrapping
    logic is delegated to the XAiWrappers class.
    """

    def register(self, client, _provider=None, stream=False):
        from memori.llm._constants import XAI_LLM_PROVIDER
        from memori.llm._xai_wrappers import XAiWrappers

        if not hasattr(client, "chat"):
            raise RuntimeError("client provided is not instance of xAI")

        if not hasattr(client, "_memori_installed"):
            if hasattr(client.chat, "completions"):
                client.beta._chat_completions_parse = client.beta.chat.completions.parse
                client.chat._completions_create = client.chat.completions.create
                client_version = getattr(client, "_version", None)

                self.config.framework.provider = _provider
                self.config.llm.provider = XAI_LLM_PROVIDER
                self.config.llm.version = client_version

                try:
                    asyncio.get_running_loop()

                    if stream is True:
                        client.beta.chat.completions.parse = (
                            InvokeAsyncStream(
                                self.config, client.beta._chat_completions_parse
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                        client.chat.completions.create = (
                            InvokeAsyncStream(
                                self.config,
                                client.chat._completions_create,
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                    else:
                        client.beta.chat.completions.parse = (
                            InvokeAsync(
                                self.config, client.beta._chat_completions_parse
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                        client.chat.completions.create = (
                            InvokeAsync(
                                self.config,
                                client.chat._completions_create,
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                except RuntimeError:
                    if stream is True:
                        client.beta.chat.completions.parse = (
                            InvokeStream(
                                self.config, client.beta._chat_completions_parse
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                        client.chat.completions.create = (
                            InvokeStream(
                                self.config,
                                client.chat._completions_create,
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                    else:
                        client.beta.chat.completions.parse = (
                            Invoke(self.config, client.beta._chat_completions_parse)
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
                        client.chat.completions.create = (
                            Invoke(
                                self.config,
                                client.chat._completions_create,
                            )
                            .set_client(_provider, XAI_LLM_PROVIDER, client_version)
                            .invoke
                        )
            else:
                client.chat._create = client.chat.create
                client_version = getattr(client, "_version", None)

                self.config.framework.provider = _provider
                self.config.llm.provider = XAI_LLM_PROVIDER
                self.config.llm.version = client_version

                wrappers = XAiWrappers(self.config)

                def wrapped_create(*args, **kwargs):
                    kwargs = wrappers.inject_conversation_history(kwargs)
                    chat_obj = client.chat._create(*args, **kwargs)
                    wrappers.wrap_chat_methods(chat_obj, client_version)
                    return chat_obj

                client.chat.create = wrapped_create

            client._memori_installed = True

        return self


class Agno(BaseClient):
    def register(self, openai_chat=None, claude=None, gemini=None, xai=None):
        if openai_chat is None and claude is None and gemini is None and xai is None:
            raise RuntimeError("Agno::register called without model")

        if openai_chat is not None:
            if not self._is_agno_openai_model(openai_chat):
                raise RuntimeError(
                    "model provided is not instance of agno.models.openai.OpenAIChat"
                )
            client = openai_chat.get_client()
            OpenAi(self.config).register(client, _provider=AGNO_FRAMEWORK_PROVIDER)

            if not hasattr(openai_chat, "_memori_original_get_client"):
                original_get_client = openai_chat.get_client
                openai_chat._memori_original_get_client = original_get_client
                openai_wrapper = OpenAi(self.config)

                def wrapped_get_client():
                    client = openai_chat._memori_original_get_client()
                    openai_wrapper.register(client, _provider=AGNO_FRAMEWORK_PROVIDER)
                    return client

                openai_chat.get_client = wrapped_get_client

                # Also wrap get_async_client for async support
                if hasattr(openai_chat, "get_async_client"):
                    original_get_async_client = openai_chat.get_async_client
                    openai_chat._memori_original_get_async_client = (
                        original_get_async_client
                    )

                    def wrapped_get_async_client():
                        client = openai_chat._memori_original_get_async_client()
                        openai_wrapper.register(
                            client, _provider=AGNO_FRAMEWORK_PROVIDER
                        )
                        return client

                    openai_chat.get_async_client = wrapped_get_async_client

        if claude is not None:
            if not self._is_agno_anthropic_model(claude):
                raise RuntimeError(
                    "model provided is not instance of agno.models.anthropic.Claude"
                )
            client = claude.get_client()
            Anthropic(self.config).register(client, _provider=AGNO_FRAMEWORK_PROVIDER)

            if not hasattr(claude, "_memori_original_get_client"):
                original_get_client = claude.get_client
                claude._memori_original_get_client = original_get_client
                anthropic_wrapper = Anthropic(self.config)

                def wrapped_get_client():
                    client = claude._memori_original_get_client()
                    anthropic_wrapper.register(
                        client, _provider=AGNO_FRAMEWORK_PROVIDER
                    )
                    return client

                claude.get_client = wrapped_get_client

                # Also wrap get_async_client for async support
                if hasattr(claude, "get_async_client"):
                    original_get_async_client = claude.get_async_client
                    claude._memori_original_get_async_client = original_get_async_client

                    def wrapped_get_async_client():
                        client = claude._memori_original_get_async_client()
                        anthropic_wrapper.register(
                            client, _provider=AGNO_FRAMEWORK_PROVIDER
                        )
                        return client

                    claude.get_async_client = wrapped_get_async_client

        if gemini is not None:
            if not self._is_agno_google_model(gemini):
                raise RuntimeError(
                    "model provided is not instance of agno.models.google.Gemini"
                )
            client = gemini.get_client()
            Google(self.config).register(client, _provider=AGNO_FRAMEWORK_PROVIDER)

            # Wrap get_client to ensure all future client instances are wrapped
            if not hasattr(gemini, "_memori_original_get_client"):
                original_get_client = gemini.get_client
                gemini._memori_original_get_client = original_get_client
                google_wrapper = Google(self.config)

                def wrapped_get_client():
                    client = gemini._memori_original_get_client()
                    google_wrapper.register(client, _provider=AGNO_FRAMEWORK_PROVIDER)
                    return client

                gemini.get_client = wrapped_get_client

        if xai is not None:
            if not self._is_agno_xai_model(xai):
                raise RuntimeError(
                    "model provided is not instance of agno.models.xai.xAI"
                )
            client = xai.get_client()
            XAi(self.config).register(client, _provider=AGNO_FRAMEWORK_PROVIDER)

            if not hasattr(xai, "_memori_original_get_client"):
                original_get_client = xai.get_client
                xai._memori_original_get_client = original_get_client
                xai_wrapper = XAi(self.config)

                def wrapped_get_client():
                    client = xai._memori_original_get_client()
                    xai_wrapper.register(client, _provider=AGNO_FRAMEWORK_PROVIDER)
                    return client

                xai.get_client = wrapped_get_client

                # Also wrap get_async_client for async support
                if hasattr(xai, "get_async_client"):
                    original_get_async_client = xai.get_async_client
                    xai._memori_original_get_async_client = original_get_async_client

                    def wrapped_get_async_client():
                        client = xai._memori_original_get_async_client()
                        xai_wrapper.register(client, _provider=AGNO_FRAMEWORK_PROVIDER)
                        return client

                    xai.get_async_client = wrapped_get_async_client

        return self

    def _is_agno_openai_model(self, model):
        return "agno.models.openai" in str(type(model).__module__)

    def _is_agno_anthropic_model(self, model):
        return "agno.models.anthropic" in str(type(model).__module__)

    def _is_agno_google_model(self, model):
        return "agno.models.google" in str(type(model).__module__)

    def _is_agno_xai_model(self, model):
        return "agno.models.xai" in str(type(model).__module__)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import json
import os
import pprint
import traceback

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from memori._config import Config


class Api:
    def __init__(self, config: Config):
        self.__base = os.environ.get("MEMORI_API_URL_BASE")
        if self.__base is None:
            self.__base = "https://api.memorilabs.ai"

        self.config = config

    def get(self, route):
        r = self.__session().get(
            self.url(route), headers={"Authorization": f"Bearer {self.config.api_key}"}
        )

        r.raise_for_status()

        return r.json()

    def patch(self, route, json=None):
        if json is None:
            json = {}
        r = self.__session().patch(
            self.url(route),
            headers={"Authorization": f"Bearer {self.config.api_key}"},
            json=json,
        )

        r.raise_for_status()

        return r.json()

    def post(self, route, json=None):
        if json is None:
            json = {}
        r = self.__session().post(
            self.url(route),
            headers={"Authorization": f"Bearer {self.config.api_key}"},
            json=json,
        )

        r.raise_for_status()

        return r.json()

    def __session(self):
        adapter = HTTPAdapter(
            max_retries=_ApiRetryRecoverable(
                allowed_methods=["GET", "PATCH", "POST", "PUT", "DELETE"],
                backoff_factor=1,
                raise_on_status=False,
                status=None,
                total=5,
            )
        )

        session = requests.Session()
        session.mount("https://", adapter)
        session.mount("http://", adapter)

        return session

    def url(self, route):
        return f"{self.__base}/v1/-/{route}"


class _ApiRetryRecoverable(Retry):
    def is_retry(self, method, status_code, has_retry_after=False):
        return 500 <= status_code <= 599


class Collector:
    def __init__(self, config: Config):
        self.__base = os.environ.get("MEMORI_COLLECTOR_URL_BASE")
        if self.__base is None:
            self.__base = "https://api.memorilabs.ai"

        self.config = config

    def fire_and_forget(self, payload):
        if not self.config.is_test_mode():
            try:
                requests.post(
                    f"{self.__base}/rec",
                    json=payload,
                    timeout=self.config.request_secs_timeout,
                )
            except Exception:
                payload["meta"]["fnfg"] = {
                    "exc": traceback.format_exc(),
                    "status": "recovered",
                }

                try:
                    requests.post(
                        f"{self.__base}/rec",
                        json=json.loads(json.dumps(payload, default=str)),
                        timeout=self.config.request_secs_timeout,
                    )
                except Exception:
                    if self.config.raise_final_request_attempt is True:
                        raise
        else:
            pprint.pprint(payload)

        return self



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.storage.cockroachdb._files import Files


class Display:
    def __init__(self):
        self.files = Files()

    def cluster_already_started(self):
        return """You already have an active CockroachDB cluster running. To start
  a new one, execute this command first:

    python -m memori cockroachdb cluster delete
"""

    def cluster_was_not_started(self):
        return """You do not have an active CockroachDB cluster running. To start
  a new one, execute this command first:

    python -m memori cockroachdb cluster start
"""



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import os


class Files:
    def __init__(self):
        self.filename_cluster_id = "id"

    def cluster_dir(self):
        return self.storage_dir() + "/cluster"

    def cluster_id(self):
        return "/".join([self.cluster_dir(), self.filename_cluster_id])

    def makedirs(self):
        os.makedirs(self.cluster_dir(), exist_ok=True)
        return self

    def read_id(self):
        try:
            with open(self.cluster_id()) as f:
                return f.read().lstrip().rstrip()
        except FileNotFoundError:
            return None

    def remove_id(self):
        try:
            os.unlink(self.cluster_id())
        except FileNotFoundError:
            pass

        return self

    def storage_dir(self):
        home = os.environ.get("MEMORI_HOME", None)
        if home is None:
            home = os.environ.get("HOME", None)

        if home is None:
            raise RuntimeError(
                "neither MEMORI_HOME nor HOME environment variable is set"
            )

        return f"{home}/.memori"

    def write_id(self, id):
        self.makedirs()

        with open(self.cluster_id(), "w") as f:
            f.write(id)

        return self



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

migrations = {
    1: [
        {
            "description": "create table memori_schema_version",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_schema_version(
                            num NUMBER(19) NOT NULL PRIMARY KEY
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_entity",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_entity(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            external_id VARCHAR2(100) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_entity_external_id UNIQUE (external_id),
                            CONSTRAINT uk_memori_entity_uuid UNIQUE (uuid)
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_process",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_process(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            external_id VARCHAR2(100) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_process_external_id UNIQUE (external_id),
                            CONSTRAINT uk_memori_process_uuid UNIQUE (uuid)
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_session",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_session(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            entity_id NUMBER(19) DEFAULT NULL,
                            process_id NUMBER(19) DEFAULT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_session_entity_id UNIQUE (entity_id, id),
                            CONSTRAINT uk_memori_session_process_id UNIQUE (process_id, id),
                            CONSTRAINT uk_memori_session_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_sess_entity
                               FOREIGN KEY (entity_id)
                                REFERENCES memori_entity (id)
                                 ON DELETE CASCADE,
                            CONSTRAINT fk_memori_sess_process
                               FOREIGN KEY (process_id)
                                REFERENCES memori_process (id)
                                 ON DELETE CASCADE
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_conversation",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_conversation(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            session_id NUMBER(19) NOT NULL,
                            summary CLOB DEFAULT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_conversation_session_id UNIQUE (session_id),
                            CONSTRAINT uk_memori_conversation_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_conv_session
                               FOREIGN KEY (session_id)
                                REFERENCES memori_session (id)
                                 ON DELETE CASCADE
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_conversation_message",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_conversation_message(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            conversation_id NUMBER(19) NOT NULL,
                            role VARCHAR2(255) NOT NULL,
                            type VARCHAR2(255) DEFAULT NULL,
                            content CLOB NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_conversation_message_conversation_id UNIQUE (conversation_id, id),
                            CONSTRAINT uk_memori_conversation_message_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_conv_msg_conv
                               FOREIGN KEY (conversation_id)
                                REFERENCES memori_conversation (id)
                                 ON DELETE CASCADE
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_entity_fact",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_entity_fact(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            entity_id NUMBER(19) NOT NULL,
                            content CLOB NOT NULL,
                            content_embedding BLOB NOT NULL,
                            num_times NUMBER(19) NOT NULL,
                            date_last_time TIMESTAMP NOT NULL,
                            uniq CHAR(64) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_entity_fact_entity_id UNIQUE (entity_id, id),
                            CONSTRAINT uk_memori_entity_fact_entity_id_uniq UNIQUE (entity_id, uniq),
                            CONSTRAINT uk_memori_entity_fact_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_ent_fact_entity
                               FOREIGN KEY (entity_id)
                                REFERENCES memori_entity (id)
                                 ON DELETE CASCADE
                        )
                    ';
                    EXECUTE IMMEDIATE '
                        CREATE INDEX idx_memori_entity_fact_entity_id_freq
                        ON memori_entity_fact (entity_id, num_times DESC, date_last_time DESC)
                    ';
                    EXECUTE IMMEDIATE '
                        CREATE INDEX idx_memori_entity_fact_embedding_search
                        ON memori_entity_fact (entity_id, id)
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 OR SQLCODE = -1408 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_process_attribute",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_process_attribute(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            process_id NUMBER(19) NOT NULL,
                            content CLOB NOT NULL,
                            num_times NUMBER(19) NOT NULL,
                            date_last_time TIMESTAMP NOT NULL,
                            uniq CHAR(64) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_process_attribute_process_id UNIQUE (process_id, id),
                            CONSTRAINT uk_memori_process_attribute_process_id_uniq UNIQUE (process_id, uniq),
                            CONSTRAINT uk_memori_process_attribute_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_proc_attribute
                               FOREIGN KEY (process_id)
                                REFERENCES memori_process (id)
                                 ON DELETE CASCADE
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_subject",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_subject(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            name VARCHAR2(255) NOT NULL,
                            type VARCHAR2(255) NOT NULL,
                            uniq CHAR(64) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_subject_uniq UNIQUE (uniq),
                            CONSTRAINT uk_memori_subject_uuid UNIQUE (uuid)
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_predicate",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_predicate(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            content CLOB NOT NULL,
                            uniq CHAR(64) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_predicate_uniq UNIQUE (uniq),
                            CONSTRAINT uk_memori_predicate_uuid UNIQUE (uuid)
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_object",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_object(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            name VARCHAR2(255) NOT NULL,
                            type VARCHAR2(255) NOT NULL,
                            uniq CHAR(64) NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_object_uniq UNIQUE (uniq),
                            CONSTRAINT uk_memori_object_uuid UNIQUE (uuid)
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
        {
            "description": "create table memori_knowledge_graph",
            "operation": """
                BEGIN
                    EXECUTE IMMEDIATE '
                        CREATE TABLE memori_knowledge_graph(
                            id NUMBER(19) GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
                            uuid VARCHAR2(36) NOT NULL,
                            entity_id NUMBER(19) NOT NULL,
                            subject_id NUMBER(19) NOT NULL,
                            predicate_id NUMBER(19) NOT NULL,
                            object_id NUMBER(19) NOT NULL,
                            num_times NUMBER(19) NOT NULL,
                            date_last_time TIMESTAMP NOT NULL,
                            date_created TIMESTAMP DEFAULT SYSTIMESTAMP NOT NULL,
                            date_updated TIMESTAMP DEFAULT NULL,
                            CONSTRAINT uk_memori_knowledge_graph_entity_id UNIQUE (entity_id, id),
                            CONSTRAINT uk_memori_knowledge_graph_entity_subject_predicate_object UNIQUE (entity_id, subject_id, predicate_id, object_id),
                            CONSTRAINT uk_memori_knowledge_graph_object_id UNIQUE (object_id, id),
                            CONSTRAINT uk_memori_knowledge_graph_predicate_id UNIQUE (predicate_id, id),
                            CONSTRAINT uk_memori_knowledge_graph_subject_id UNIQUE (subject_id, id),
                            CONSTRAINT uk_memori_knowledge_graph_uuid UNIQUE (uuid),
                            CONSTRAINT fk_memori_know_graph_entity
                               FOREIGN KEY (entity_id)
                                REFERENCES memori_entity (id)
                                 ON DELETE CASCADE,
                            CONSTRAINT fk_memori_know_graph_object
                               FOREIGN KEY (object_id)
                                REFERENCES memori_object (id)
                                 ON DELETE CASCADE,
                            CONSTRAINT fk_memori_know_graph_predicate
                               FOREIGN KEY (predicate_id)
                                REFERENCES memori_predicate (id)
                                 ON DELETE CASCADE,
                            CONSTRAINT fk_memori_know_graph_subject
                               FOREIGN KEY (subject_id)
                                REFERENCES memori_subject (id)
                                 ON DELETE CASCADE
                        )
                    ';
                EXCEPTION
                    WHEN OTHERS THEN
                        IF SQLCODE = -955 THEN NULL;
                        ELSE RAISE;
                        END IF;
                END;
            """,
        },
    ]
}



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

migrations = {
    1: [
        {
            "description": "create collection memori_schema_version",
            "operations": [
                {
                    "collection": "memori_schema_version",
                    "method": "create_index",
                    "args": [[("num", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_entity",
            "operations": [
                {
                    "collection": "memori_entity",
                    "method": "create_index",
                    "args": [[("external_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_entity",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_process",
            "operations": [
                {
                    "collection": "memori_process",
                    "method": "create_index",
                    "args": [[("external_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_process",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_session",
            "operations": [
                {
                    "collection": "memori_session",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_session",
                    "method": "create_index",
                    "args": [[("entity_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True, "sparse": True},
                },
                {
                    "collection": "memori_session",
                    "method": "create_index",
                    "args": [[("process_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True, "sparse": True},
                },
            ],
        },
        {
            "description": "create collection memori_conversation",
            "operations": [
                {
                    "collection": "memori_conversation",
                    "method": "create_index",
                    "args": [[("session_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_conversation",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_conversation_message",
            "operations": [
                {
                    "collection": "memori_conversation_message",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_conversation_message",
                    "method": "create_index",
                    "args": [[("conversation_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_entity_fact",
            "operations": [
                {
                    "collection": "memori_entity_fact",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_entity_fact",
                    "method": "create_index",
                    "args": [[("entity_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_entity_fact",
                    "method": "create_index",
                    "args": [[("entity_id", 1), ("uniq", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_entity_fact",
                    "method": "create_index",
                    "args": [
                        [("entity_id", 1), ("num_times", -1), ("date_last_time", -1)]
                    ],
                    "kwargs": {"name": "idx_memori_entity_fact_entity_id_freq"},
                },
                {
                    "collection": "memori_entity_fact",
                    "method": "create_index",
                    "args": [[("entity_id", 1), ("_id", 1)]],
                    "kwargs": {"name": "idx_memori_entity_fact_embedding_search"},
                },
            ],
        },
        {
            "description": "create collection memori_process_attribute",
            "operations": [
                {
                    "collection": "memori_process_attribute",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_process_attribute",
                    "method": "create_index",
                    "args": [[("process_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_process_attribute",
                    "method": "create_index",
                    "args": [[("process_id", 1), ("uniq", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_subject",
            "operations": [
                {
                    "collection": "memori_subject",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_subject",
                    "method": "create_index",
                    "args": [[("uniq", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_predicate",
            "operations": [
                {
                    "collection": "memori_predicate",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_predicate",
                    "method": "create_index",
                    "args": [[("uniq", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_object",
            "operations": [
                {
                    "collection": "memori_object",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_object",
                    "method": "create_index",
                    "args": [[("uniq", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
        {
            "description": "create collection memori_knowledge_graph",
            "operations": [
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [[("uuid", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [[("entity_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [
                        [
                            ("entity_id", 1),
                            ("subject_id", 1),
                            ("predicate_id", 1),
                            ("object_id", 1),
                        ]
                    ],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [[("subject_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [[("predicate_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
                {
                    "collection": "memori_knowledge_graph",
                    "method": "create_index",
                    "args": [[("object_id", 1), ("_id", 1)]],
                    "kwargs": {"unique": True},
                },
            ],
        },
    ]
}



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.memory.augmentation.augmentations.memori._augmentation import (
    AdvancedAugmentation,
)

__all__ = ["AdvancedAugmentation"]



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from collections.abc import Callable
from typing import Any

from memori.storage._base import BaseStorageAdapter


class Registry:
    _adapters: dict[Callable[[Any], bool], type[BaseStorageAdapter]] = {}
    _drivers: dict[str, type] = {}

    @classmethod
    def register_adapter(cls, matcher: Callable[[Any], bool]):
        def decorator(adapter_class: type[BaseStorageAdapter]):
            cls._adapters[matcher] = adapter_class
            return adapter_class

        return decorator

    @classmethod
    def register_driver(cls, dialect: str):
        def decorator(driver_class: type):
            cls._drivers[dialect] = driver_class
            return driver_class

        return decorator

    def adapter(self, conn: Any) -> BaseStorageAdapter:
        conn_to_check = conn() if callable(conn) else conn

        for matcher, adapter_class in self._adapters.items():
            if matcher(conn_to_check):
                return adapter_class(lambda: conn_to_check)

        raise RuntimeError(
            f"Unsupported database: {type(conn_to_check).__module__}.{type(conn_to_check).__name__}"
        )

    def driver(self, conn: BaseStorageAdapter):
        dialect = conn.get_dialect()
        if dialect not in self._drivers:
            raise RuntimeError(f"Unsupported database dialect: {dialect}")
        return self._drivers[dialect](conn)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori._cli import Cli
from memori._config import Config
from memori._network import Api


class Manager:
    def __init__(self, config: Config):
        self.config = config

    def execute(self):
        cli = Cli(self.config)

        response = Api(self.config).get("sdk/quota")

        cli.notice("Maximum # of Memories: " + f"{response['memories']['max']:,}")
        cli.notice("Current # of Memories: " + f"{response['memories']['num']:,}\n")
        cli.notice(f"{response['message']}\n")



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import os
from concurrent.futures import ThreadPoolExecutor


class Cache:
    def __init__(self):
        self.conversation_id = None
        self.entity_id = None
        self.process_id = None
        self.session_id = None


class Storage:
    def __init__(self):
        self.cockroachdb = False


class Config:
    def __init__(self):
        self.api_key = None
        self.augmentation = None
        self.cache = Cache()
        self.enterprise = False
        self.llm = Llm()
        self.framework = Framework()
        self.entity_id = None
        self.process_id = None
        self.raise_final_request_attempt = True
        self.recall_embeddings_limit = 1000
        self.recall_facts_limit = 5
        self.recall_relevance_threshold = 0.1
        self.request_backoff_factor = 1
        self.request_num_backoff = 5
        self.request_secs_timeout = 5
        self.session_id = None
        self.session_timeout_minutes = 30
        self.storage = None
        self.storage_config = Storage()
        self.thread_pool_executor = ThreadPoolExecutor(max_workers=15)
        self.version = "3.0.3"

    def is_test_mode(self):
        return os.environ.get("MEMORI_TEST_MODE", None) is not None

    def reset_cache(self):
        self.cache = Cache()
        return self


class Framework:
    def __init__(self):
        self.provider = None


class Llm:
    def __init__(self):
        self.provider = None
        self.version = None



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.storage._base import BaseStorageAdapter
from memori.storage._registry import Registry


class CursorWrapper:
    def __init__(self, cursor):
        self._cursor = cursor

    def mappings(self):
        return MappingResult(self._cursor)

    def __getattr__(self, name):
        return getattr(self._cursor, name)


class MappingResult:
    def __init__(self, cursor):
        self._cursor = cursor

    def fetchone(self):
        row = self._cursor.fetchone()
        if row is None:
            return None
        columns = [col[0] for col in self._cursor.description]
        return dict(zip(columns, row, strict=True))

    def fetchall(self):
        rows = self._cursor.fetchall()
        columns = [col[0] for col in self._cursor.description]
        return [dict(zip(columns, row, strict=True)) for row in rows]


def is_django_connection(conn):
    if not hasattr(conn, "__class__"):
        return False

    module_name = conn.__class__.__module__
    if not module_name.startswith("django.db"):
        return False

    if not (hasattr(conn, "cursor") and callable(getattr(conn, "cursor", None))):
        return False

    return True


@Registry.register_adapter(is_django_connection)
class Adapter(BaseStorageAdapter):
    def commit(self):
        self.conn.commit()
        return self

    def execute(self, operation, binds=()):
        cursor = self.conn.cursor()
        try:
            cursor.execute(operation, binds)
            return CursorWrapper(cursor)
        except Exception:
            cursor.close()
            raise

    def flush(self):
        return self

    def get_dialect(self):
        vendor = self.conn.vendor
        dialect_mapping = {
            "postgresql": "postgresql",
            "mysql": "mysql",
            "sqlite": "sqlite",
            "oracle": "oracle",
        }
        if vendor in dialect_mapping:
            return dialect_mapping[vendor]
        raise ValueError(f"Unable to determine dialect from Django vendor: {vendor}")

    def rollback(self):
        self.conn.rollback()
        return self



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from collections.abc import Callable
from typing import Any

from memori.llm._base import BaseClient, BaseLlmAdaptor


class Registry:
    _clients: dict[Callable[[Any], bool], type[BaseClient]] = {}
    _adapters: dict[Callable[[str | None, str | None], bool], type[BaseLlmAdaptor]] = {}

    @classmethod
    def register_client(cls, matcher: Callable[[Any], bool]):
        def decorator(client_class: type[BaseClient]):
            cls._clients[matcher] = client_class
            return client_class

        return decorator

    @classmethod
    def register_adapter(cls, matcher: Callable[[str | None, str | None], bool]):
        def decorator(adapter_class: type[BaseLlmAdaptor]):
            cls._adapters[matcher] = adapter_class
            return adapter_class

        return decorator

    def client(self, client_obj: Any, config) -> BaseClient:
        for matcher, client_class in self._clients.items():
            if matcher(client_obj):
                return client_class(config)

        raise RuntimeError(
            f"Unsupported LLM client type: {type(client_obj).__module__}.{type(client_obj).__name__}"
        )

    def adapter(self, provider: str | None, title: str | None) -> BaseLlmAdaptor:
        for matcher, adapter_class in self._adapters.items():
            if matcher(provider, title):
                return adapter_class()

        raise RuntimeError(
            f"Unsupported LLM provider: framework={provider}, llm={title}"
        )



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import inspect
import time
from collections.abc import AsyncIterator, Iterator

from botocore.eventstream import EventStream
from grpc.experimental.aio import UnaryStreamCall

from memori._utils import merge_chunk
from memori.llm._base import BaseInvoke
from memori.llm._iterable import Iterable as MemoriIterable
from memori.llm._iterator import AsyncIterator as MemoriAsyncIterator
from memori.llm._iterator import Iterator as MemoriIterator
from memori.llm._streaming import StreamingBody as MemoriStreamingBody
from memori.llm._utils import client_is_bedrock


class Invoke(BaseInvoke):
    def invoke(self, **kwargs):
        start = time.time()

        kwargs = self.inject_conversation_messages(
            self.inject_recalled_facts(self.configure_for_streaming_usage(kwargs))
        )

        raw_response = self._method(**kwargs)

        if isinstance(raw_response, Iterator) or inspect.isgenerator(raw_response):
            return (
                MemoriIterator(self.config, raw_response)
                .configure_invoke(self)
                .configure_request(kwargs, start)
            )
        elif client_is_bedrock(
            self.config.framework.provider, self.config.llm.provider
        ):
            if isinstance(raw_response["body"], EventStream):
                raw_response["body"] = (
                    MemoriIterable(self.config, raw_response["body"])
                    .configure_invoke(self)
                    .configure_request(kwargs, start)
                )
            else:
                raw_response["body"] = (
                    MemoriStreamingBody(self.config, raw_response["body"])
                    .configure_invoke(self)
                    .configure_request(kwargs, start)
                )

            return raw_response
        else:
            self.handle_post_response(kwargs, start, raw_response)
            return raw_response


class InvokeAsync(BaseInvoke):
    async def invoke(self, **kwargs):
        start = time.time()

        kwargs = self.inject_conversation_messages(
            self.inject_recalled_facts(self.configure_for_streaming_usage(kwargs))
        )

        raw_response = await self._method(**kwargs)
        self.handle_post_response(kwargs, start, raw_response)
        return raw_response


class InvokeAsyncIterator(BaseInvoke):
    async def invoke(self, **kwargs):
        start = time.time()

        kwargs = self.inject_conversation_messages(
            self.inject_recalled_facts(self.configure_for_streaming_usage(kwargs))
        )

        raw_response = await self._method(**kwargs)
        if isinstance(raw_response, AsyncIterator) or isinstance(
            raw_response, UnaryStreamCall
        ):
            return (
                MemoriAsyncIterator(self.config, raw_response)
                .configure_invoke(self)
                .configure_request(kwargs, start)
            )
        else:
            self.handle_post_response(kwargs, start, raw_response)
            return raw_response


class InvokeAsyncStream(BaseInvoke):
    async def invoke(self, **kwargs):
        start = time.time()

        kwargs = self.inject_conversation_messages(
            self.inject_recalled_facts(self.configure_for_streaming_usage(kwargs))
        )

        stream = await self._method(**kwargs)

        raw_response = {}
        async for chunk in stream:
            raw_response = merge_chunk(raw_response, chunk.__dict__)
            yield chunk

        self.handle_post_response(kwargs, start, raw_response)


class InvokeStream(BaseInvoke):
    async def invoke(self, **kwargs):
        start = time.time()

        kwargs = self.inject_conversation_messages(
            self.inject_recalled_facts(self.configure_for_streaming_usage(kwargs))
        )

        raw_response = await self._method(**kwargs)

        self.handle_post_response(kwargs, start, raw_response)
        return raw_response



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.llm import _clients  # noqa: F401
from memori.llm._registry import Registry
from memori.llm.adapters import anthropic, bedrock, google, openai, xai  # noqa: F401

__all__ = ["Registry"]



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.memory.augmentation.input import AugmentationInput


class AugmentationContext:
    def __init__(self, payload: AugmentationInput):
        self.payload = payload
        self.data = {}
        self.writes = []

    def add_write(self, method_path: str, *args, **kwargs):
        self.writes.append({"method_path": method_path, "args": args, "kwargs": kwargs})
        return self


class BaseAugmentation:
    def __init__(self, config=None, enabled: bool = True):
        self.config = config
        self.enabled = enabled

    async def process(self, ctx: AugmentationContext, driver) -> AugmentationContext:
        raise NotImplementedError("Augmentation must implement process() method")



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                 perfectam memoriam
                      memorilabs.ai
"""

import asyncio
import os
import struct
from typing import Any

os.environ.setdefault("TOKENIZERS_PARALLELISM", "false")

from sentence_transformers import SentenceTransformer

_MODEL_CACHE: dict[str, SentenceTransformer] = {}
_DEFAULT_DIMENSION = 768


def _get_model(model_name: str) -> SentenceTransformer:
    if model_name not in _MODEL_CACHE:
        _MODEL_CACHE[model_name] = SentenceTransformer(model_name)
    return _MODEL_CACHE[model_name]


def format_embedding_for_db(embedding: list[float], dialect: str) -> Any:
    """Format embedding for database storage.

    Args:
        embedding: List of floats representing the embedding vector
        dialect: Database dialect (postgresql, mysql, sqlite, mongodb)

    Returns:
        Formatted embedding optimized for the target database:
        - PostgreSQL/CockroachDB/MySQL/SQLite: Binary (BYTEA/BLOB) - compact & fast
        - MongoDB: Binary (BinData) - compact & fast
    """
    binary_data = struct.pack(f"<{len(embedding)}f", *embedding)

    if dialect == "mongodb":
        try:
            import bson

            return bson.Binary(binary_data)
        except ImportError:
            return binary_data
    else:
        return binary_data


def embed_texts(
    texts: str | list[str], model: str = "all-mpnet-base-v2"
) -> list[list[float]]:
    inputs = [texts] if isinstance(texts, str) else [t for t in texts if t]
    if not inputs:
        return []

    try:
        encoder = _get_model(model)
    except (OSError, RuntimeError, ValueError):
        return [[0.0] * _DEFAULT_DIMENSION for _ in inputs]

    try:
        embeddings = encoder.encode(inputs, convert_to_numpy=True)
        return embeddings.tolist()
    except (RuntimeError, ValueError):
        try:
            dim_value = encoder.get_sentence_embedding_dimension()
            dim = int(dim_value) if dim_value is not None else _DEFAULT_DIMENSION
        except (RuntimeError, ValueError, AttributeError, TypeError):
            dim = _DEFAULT_DIMENSION
        return [[0.0] * dim for _ in inputs]


async def embed_texts_async(
    texts: str | list[str], model: str = "all-mpnet-base-v2"
) -> list[list[float]]:
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, embed_texts, texts, model)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import os
import sys
import time
from typing import Any

import psycopg
import requests

from memori._cli import Cli
from memori._config import Config
from memori._network import Api
from memori.storage._builder import Builder
from memori.storage._manager import Manager
from memori.storage.cockroachdb._display import Display
from memori.storage.cockroachdb._files import Files


class ClusterManager:
    def __init__(self, config: Config):
        self.config = config
        self.display = Display()
        self.files = Files()

    def claim(self, cli: Cli):
        if not self.cluster_is_started():
            cli.notice(self.display.cluster_was_not_started())
            return self

        cluster_id = self.files.read_id()
        if cluster_id is not None:
            claim = Api(Config()).get(f"cockroachdb/cluster/claim/{cluster_id}")

        cli.notice(
            "You can claim the CockroachDB cluster by using this URL:\n    "
            + claim["url"]
        )
        cli.newline()

        return self

    def delete(self, cli: Cli):
        if not self.cluster_is_started():
            cli.notice(self.display.cluster_was_not_started())
            return self

        cluster_id = self.files.read_id()
        if cluster_id is not None:
            try:
                Api(Config()).delete(f"cockroachdb/cluster/{cluster_id}")
            except requests.exceptions.HTTPError as e:
                if e.response is None or e.response.status_code != 404:
                    raise

        self.files.remove_id()

        cli.notice("The CockroachDB cluster has been deleted.")
        cli.newline()

        return self

    def execute(self):
        cli = Cli(self.config)

        if sys.argv[2] != "cluster" or sys.argv[3] not in ["start", "claim", "delete"]:
            self.usage()
            cli.newline()
            sys.exit(1)

        if sys.argv[3] == "claim":
            self.claim(cli)
        elif sys.argv[3] == "start":
            self.start(cli)
        elif sys.argv[3] == "delete":
            self.delete(cli)

        return self

    def start(self, cli: Cli):
        if self.cluster_is_started():
            cli.notice(self.display.cluster_already_started())
            return self

        cli.print(
            "Before we begin, we want you to know that security and privacy are\n"
            + "very important to us. Once we create this cluster for you we can\n"
            + "never access it again unless you provide us with your connection\n"
            + "string and we cannot help you recover your credentials if you lose\n"
            + "them.\n"
        )

        cli.print("This process may take a minute or longer, please be patient.")

        cli.newline()
        proceed = input("Proceed [Y/n] ")
        cli.newline()

        if proceed is not None and proceed not in ["y", "Y", ""]:
            sys.exit(0)

        cli.notice("[Step 1] Starting new cluster: ", end="")

        started = Api(Config()).post("cockroachdb/cluster/start")

        cli.print("done")

        cli.notice("[Step 2] Waiting for cluster to come online: ", end="")

        finalized: dict[str, Any] | None = None
        for _i in range(24):
            finalized = Api(Config()).post(
                f"cockroachdb/cluster/finalize/{started['cluster']['uuid']}",
                json={"cluster": {"id": started["cluster"]["id"]}},
            )
            if finalized["status"] == 1:
                break

            time.sleep(5)

        if finalized is None or finalized["status"] != 1:
            self.cluster_finalize_failed()
            return

        self.files.write_id(started["cluster"]["uuid"])

        cli.print("done")

        cli.notice("[Step 3] Creating the Memori schema:\n")

        self.config.storage = Manager(self.config).start(
            lambda: psycopg.connect(finalized["connection"]["string"])
        )
        Builder(self.config).disable_banner().execute()

        cli.notice("--- YOUR COCKROACHDB CLUSTER IS READY! ---\n")

        cli.notice("Please save your connection string:")
        cli.notice(finalized["connection"]["string"], 1)
        cli.newline()

        cli.notice("To use your cluster, set the following environment variable:")
        cli.notice(
            f"MEMORI_COCKROACHDB_CONNECTION_STRING={finalized['connection']['string']}",
            1,
        )
        cli.newline()

        cli.notice("You have to claim this database in 7 days or it will be deleted!")
        cli.notice(finalized["claim"]["url"], 1)

        cli.newline()

        cli.notice("You're all set!")
        cli.newline()

    def cluster_finalize_failed(self):
        raise RuntimeError("the cluster failed to come online; please try again")

    def cluster_is_started(self):
        return os.path.isfile(self.files.cluster_id())

    def usage(self):
        print("usage: python -m memori cockroachdb cluster <start | claim | delete>")



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from uuid import uuid4

from memori.storage._base import (
    BaseConversation,
    BaseConversationMessage,
    BaseConversationMessages,
    BaseEntity,
    BaseEntityFact,
    BaseKnowledgeGraph,
    BaseProcess,
    BaseProcessAttribute,
    BaseSchema,
    BaseSchemaVersion,
    BaseSession,
    BaseStorageAdapter,
)
from memori.storage._registry import Registry
from memori.storage.migrations._oracle import migrations


class Conversation(BaseConversation):
    def __init__(self, conn: BaseStorageAdapter):
        super().__init__(conn)
        self.message = ConversationMessage(conn)
        self.messages = ConversationMessages(conn)

    def create(self, session_id, timeout_minutes: int):
        existing = (
            self.conn.execute(
                """
                SELECT c.id,
                       COALESCE(MAX(m.date_created), c.date_created) as last_activity
                  FROM memori_conversation c
                  LEFT JOIN memori_conversation_message m ON m.conversation_id = c.id
                 WHERE c.session_id = :1
                 GROUP BY c.id, c.date_created
                """,
                (session_id,),
            )
            .mappings()
            .fetchone()
        )

        if existing:
            result = self.conn.execute(
                """
                SELECT ROUND((CAST(SYSTIMESTAMP AS DATE) - CAST(:1 AS DATE)) * 24 * 60) as minutes_since_activity
                  FROM DUAL
                """,
                (existing["last_activity"],),
            ).fetchone()

            if result and result[0] is not None and result[0] <= timeout_minutes:
                return existing["id"]

        uuid = str(uuid4())
        self.conn.execute(
            """
            MERGE INTO memori_conversation dst
            USING (SELECT :1 AS uuid, :2 AS session_id FROM DUAL) src
            ON (dst.session_id = src.session_id)
            WHEN NOT MATCHED THEN
                INSERT (uuid, session_id)
                VALUES (src.uuid, src.session_id)
            """,
            (
                uuid,
                session_id,
            ),
        )
        self.conn.commit()

        return (
            self.conn.execute(
                """
                SELECT id
                  FROM memori_conversation
                 WHERE session_id = :1
                """,
                (session_id,),
            )
            .mappings()
            .fetchone()
            .get("id", None)
        )

    def update(self, id: int, summary: str):
        if summary is None:
            return self

        self.conn.execute(
            """
            UPDATE memori_conversation
               SET summary = :1
             WHERE id = :2
            """,
            (
                summary,
                id,
            ),
        )
        self.conn.commit()

        return self

    def read(self, id: int) -> dict | None:
        result = (
            self.conn.execute(
                """
                SELECT id, uuid, session_id, summary, date_created, date_updated
                  FROM memori_conversation
                 WHERE id = :1
                """,
                (id,),
            )
            .mappings()
            .fetchone()
        )

        if result is None:
            return None

        return dict(result)


class ConversationMessage(BaseConversationMessage):
    def create(self, conversation_id: int, role: str, type: str, content: str):
        self.conn.execute(
            """
            INSERT INTO memori_conversation_message(
                uuid,
                conversation_id,
                role,
                type,
                content
            ) VALUES (
                :1,
                :2,
                :3,
                :4,
                :5
            )
            """,
            (
                str(uuid4()),
                conversation_id,
                role,
                type,
                content,
            ),
        )


class ConversationMessages(BaseConversationMessages):
    def read(self, conversation_id: int):
        results = (
            self.conn.execute(
                """
                SELECT role,
                       content
                  FROM memori_conversation_message
                 WHERE conversation_id = :1
                 ORDER BY id
                """,
                (conversation_id,),
            )
            .mappings()
            .fetchall()
        )

        messages = []
        for result in results:
            messages.append({"content": result["content"], "role": result["role"]})

        return messages


class Entity(BaseEntity):
    def create(self, external_id: str):
        self.conn.execute(
            """
            MERGE INTO memori_entity dst
            USING (SELECT :1 AS uuid, :2 AS external_id FROM DUAL) src
            ON (dst.external_id = src.external_id)
            WHEN NOT MATCHED THEN
                INSERT (uuid, external_id)
                VALUES (src.uuid, src.external_id)
            """,
            (str(uuid4()), external_id),
        )
        self.conn.commit()

        return (
            self.conn.execute(
                """
                SELECT id
                  FROM memori_entity
                 WHERE external_id = :1
                """,
                (external_id,),
            )
            .mappings()
            .fetchone()
            .get("id", None)
        )


class EntityFact(BaseEntityFact):
    def create(self, entity_id: int, facts: list, fact_embeddings: list | None = None):
        if facts is None or len(facts) == 0:
            return self

        from memori._utils import generate_uniq
        from memori.llm._embeddings import format_embedding_for_db

        dialect = self.conn.get_dialect()

        for i, fact in enumerate(facts):
            embedding = (
                fact_embeddings[i]
                if fact_embeddings and i < len(fact_embeddings)
                else []
            )
            embedding_formatted = format_embedding_for_db(embedding, dialect)
            uniq = generate_uniq([fact])

            self.conn.execute(
                """
                MERGE INTO memori_entity_fact dst
                USING (SELECT :1 AS uuid, :2 AS entity_id, :3 AS content,
                              :4 AS content_embedding, :5 AS uniq FROM DUAL) src
                ON (dst.entity_id = src.entity_id AND dst.uniq = src.uniq)
                WHEN MATCHED THEN
                    UPDATE SET num_times = dst.num_times + 1,
                               date_last_time = SYSTIMESTAMP
                WHEN NOT MATCHED THEN
                    INSERT (uuid, entity_id, content, content_embedding,
                            num_times, date_last_time, uniq)
                    VALUES (src.uuid, src.entity_id, src.content, src.content_embedding,
                            1, SYSTIMESTAMP, src.uniq)
                """,
                (
                    str(uuid4()),
                    entity_id,
                    fact,
                    embedding_formatted,
                    uniq,
                ),
            )

        self.conn.commit()
        return self

    def get_embeddings(self, entity_id: int, limit: int = 1000):
        return (
            self.conn.execute(
                """
                SELECT id,
                       content_embedding
                  FROM memori_entity_fact
                 WHERE entity_id = :1
                   AND ROWNUM <= :2
                """,
                (entity_id, limit),
            )
            .mappings()
            .fetchall()
        )

    def get_facts_by_ids(self, fact_ids: list[int]):
        if not fact_ids:
            return []

        # Oracle doesn't support ANY, so we need to use IN with placeholders
        placeholders = ",".join([f":{i + 1}" for i in range(len(fact_ids))])
        query = f"""
            SELECT id,
                   content
              FROM memori_entity_fact
             WHERE id IN ({placeholders})
        """

        return self.conn.execute(query, tuple(fact_ids)).mappings().fetchall()


class KnowledgeGraph(BaseKnowledgeGraph):
    def create(self, entity_id: int, semantic_triples: list):
        if semantic_triples is None or len(semantic_triples) == 0:
            return self

        from memori._utils import generate_uniq

        for semantic_triple in semantic_triples:
            uniq = generate_uniq(
                [semantic_triple.subject_name, semantic_triple.subject_type]
            )

            self.conn.execute(
                """
                MERGE INTO memori_subject dst
                USING (SELECT :1 AS uuid, :2 AS name, :3 AS type, :4 AS uniq FROM DUAL) src
                ON (dst.uniq = src.uniq)
                WHEN NOT MATCHED THEN
                    INSERT (uuid, name, type, uniq)
                    VALUES (src.uuid, src.name, src.type, src.uniq)
                """,
                (
                    str(uuid4()),
                    semantic_triple.subject_name,
                    semantic_triple.subject_type,
                    uniq,
                ),
            )
            self.conn.commit()

            subject_id = (
                self.conn.execute(
                    """
                    SELECT id
                      FROM memori_subject
                     WHERE uniq = :1
                    """,
                    (uniq,),
                )
                .mappings()
                .fetchone()
                .get("id", None)
            )

            uniq = generate_uniq([semantic_triple.predicate])

            self.conn.execute(
                """
                MERGE INTO memori_predicate dst
                USING (SELECT :1 AS uuid, :2 AS content, :3 AS uniq FROM DUAL) src
                ON (dst.uniq = src.uniq)
                WHEN NOT MATCHED THEN
                    INSERT (uuid, content, uniq)
                    VALUES (src.uuid, src.content, src.uniq)
                """,
                (
                    str(uuid4()),
                    semantic_triple.predicate,
                    uniq,
                ),
            )
            self.conn.commit()

            predicate_id = (
                self.conn.execute(
                    """
                    SELECT id
                      FROM memori_predicate
                     WHERE uniq = :1
                    """,
                    (uniq,),
                )
                .mappings()
                .fetchone()
                .get("id", None)
            )

            uniq = generate_uniq(
                [semantic_triple.object_name, semantic_triple.object_type]
            )

            self.conn.execute(
                """
                MERGE INTO memori_object dst
                USING (SELECT :1 AS uuid, :2 AS name, :3 AS type, :4 AS uniq FROM DUAL) src
                ON (dst.uniq = src.uniq)
                WHEN NOT MATCHED THEN
                    INSERT (uuid, name, type, uniq)
                    VALUES (src.uuid, src.name, src.type, src.uniq)
                """,
                (
                    str(uuid4()),
                    semantic_triple.object_name,
                    semantic_triple.object_type,
                    uniq,
                ),
            )
            self.conn.commit()

            object_id = (
                self.conn.execute(
                    """
                    SELECT id
                      FROM memori_object
                     WHERE uniq = :1
                    """,
                    (uniq,),
                )
                .mappings()
                .fetchone()
                .get("id", None)
            )

            if (
                entity_id is not None
                and subject_id is not None
                and predicate_id is not None
                and object_id is not None
            ):
                self.conn.execute(
                    """
                    MERGE INTO memori_knowledge_graph dst
                    USING (SELECT :1 AS uuid, :2 AS entity_id, :3 AS subject_id,
                                  :4 AS predicate_id, :5 AS object_id FROM DUAL) src
                    ON (dst.entity_id = src.entity_id AND dst.subject_id = src.subject_id
                        AND dst.predicate_id = src.predicate_id AND dst.object_id = src.object_id)
                    WHEN MATCHED THEN
                        UPDATE SET num_times = dst.num_times + 1,
                                   date_last_time = SYSTIMESTAMP
                    WHEN NOT MATCHED THEN
                        INSERT (uuid, entity_id, subject_id, predicate_id, object_id,
                                num_times, date_last_time)
                        VALUES (src.uuid, src.entity_id, src.subject_id, src.predicate_id,
                                src.object_id, 1, SYSTIMESTAMP)
                    """,
                    (str(uuid4()), entity_id, subject_id, predicate_id, object_id),
                )
                self.conn.commit()

        return self


class Process(BaseProcess):
    def create(self, external_id: str):
        self.conn.execute(
            """
            MERGE INTO memori_process dst
            USING (SELECT :1 AS uuid, :2 AS external_id FROM DUAL) src
            ON (dst.external_id = src.external_id)
            WHEN NOT MATCHED THEN
                INSERT (uuid, external_id)
                VALUES (src.uuid, src.external_id)
            """,
            (str(uuid4()), external_id),
        )
        self.conn.commit()

        return (
            self.conn.execute(
                """
                SELECT id
                  FROM memori_process
                 WHERE external_id = :1
                """,
                (external_id,),
            )
            .mappings()
            .fetchone()
            .get("id", None)
        )


class ProcessAttribute(BaseProcessAttribute):
    def create(self, process_id: int, attributes: list):
        if attributes is None or len(attributes) == 0:
            return self

        from memori._utils import generate_uniq

        for attribute in attributes:
            uniq = generate_uniq([attribute])

            self.conn.execute(
                """
                MERGE INTO memori_process_attribute dst
                USING (SELECT :1 AS uuid, :2 AS process_id, :3 AS content, :4 AS uniq FROM DUAL) src
                ON (dst.process_id = src.process_id AND dst.uniq = src.uniq)
                WHEN MATCHED THEN
                    UPDATE SET num_times = dst.num_times + 1,
                               date_last_time = SYSTIMESTAMP
                WHEN NOT MATCHED THEN
                    INSERT (uuid, process_id, content, num_times, date_last_time, uniq)
                    VALUES (src.uuid, src.process_id, src.content, 1, SYSTIMESTAMP, src.uniq)
                """,
                (
                    str(uuid4()),
                    process_id,
                    attribute,
                    uniq,
                ),
            )

        self.conn.commit()
        return self


class Session(BaseSession):
    def create(self, uuid: str, entity_id: int, process_id: int):
        self.conn.execute(
            """
            MERGE INTO memori_session dst
            USING (SELECT :1 AS uuid, :2 AS entity_id, :3 AS process_id FROM DUAL) src
            ON (dst.uuid = src.uuid)
            WHEN NOT MATCHED THEN
                INSERT (uuid, entity_id, process_id)
                VALUES (src.uuid, src.entity_id, src.process_id)
            """,
            (str(uuid), entity_id, process_id),
        )
        self.conn.commit()

        return (
            self.conn.execute(
                """
                SELECT id
                  FROM memori_session
                 WHERE uuid = :1
                """,
                (str(uuid),),
            )
            .mappings()
            .fetchone()
            .get("id", None)
        )


class Schema(BaseSchema):
    def __init__(self, conn: BaseStorageAdapter):
        super().__init__(conn)
        self.version = SchemaVersion(conn)


class SchemaVersion(BaseSchemaVersion):
    def create(self, num: int):
        self.conn.execute(
            """
            INSERT INTO memori_schema_version(
                num
            ) VALUES (
                :1
            )
            """,
            (num,),
        )

    def delete(self):
        self.conn.execute(
            """
            DELETE FROM memori_schema_version
            """
        )

    def read(self):
        return (
            self.conn.execute(
                """
                SELECT num
                  FROM memori_schema_version
                """
            )
            .mappings()
            .fetchone()
            .get("num", None)
        )


@Registry.register_driver("oracle")
class Driver:
    """Oracle storage driver.

    Attributes:
        migrations: Database schema migrations for Oracle.
        requires_rollback_on_error: Oracle aborts transactions when a query
            fails and requires an explicit ROLLBACK before executing new queries.
    """

    migrations = migrations
    requires_rollback_on_error = True

    def __init__(self, conn: BaseStorageAdapter):
        self.conversation = Conversation(conn)
        self.entity = Entity(conn)
        self.entity_fact = EntityFact(conn)
        self.knowledge_graph = KnowledgeGraph(conn)
        self.process = Process(conn)
        self.process_attribute = ProcessAttribute(conn)
        self.schema = Schema(conn)
        self.session = Session(conn)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

migrations = {
    1: [
        {
            "description": "create table memori_schema_version",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_schema_version(
                    num BIGINT NOT NULL PRIMARY KEY
                )
            """,
        },
        {
            "description": "create table memori_entity",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_entity(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    external_id VARCHAR(100) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_entity_external_id UNIQUE (external_id),
                    CONSTRAINT uk_memori_entity_uuid UNIQUE (uuid)
                )
            """,
        },
        {
            "description": "create table memori_process",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_process(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    external_id VARCHAR(100) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_process_external_id UNIQUE (external_id),
                    CONSTRAINT uk_memori_process_uuid UNIQUE (uuid)
                )
            """,
        },
        {
            "description": "create table memori_session",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_session(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    entity_id BIGINT DEFAULT NULL,
                    process_id BIGINT DEFAULT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_session_entity_id UNIQUE (entity_id, id),
                    CONSTRAINT uk_memori_session_process_id UNIQUE (process_id, id),
                    CONSTRAINT uk_memori_session_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_sess_entity
                       FOREIGN KEY (entity_id)
                        REFERENCES memori_entity (id)
                         ON DELETE CASCADE,
                    CONSTRAINT fk_memori_sess_process
                       FOREIGN KEY (process_id)
                        REFERENCES memori_process (id)
                         ON DELETE CASCADE
                )
            """,
        },
        {
            "description": "create table memori_conversation",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_conversation(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    session_id BIGINT NOT NULL,
                    summary TEXT DEFAULT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_conversation_session_id UNIQUE (session_id),
                    CONSTRAINT uk_memori_conversation_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_conv_session
                       FOREIGN KEY (session_id)
                        REFERENCES memori_session (id)
                         ON DELETE CASCADE
                )
            """,
        },
        {
            "description": "create table memori_conversation_message",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_conversation_message(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    conversation_id BIGINT NOT NULL,
                    role VARCHAR(255) NOT NULL,
                    type VARCHAR(255) DEFAULT NULL,
                    content TEXT NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_conversation_message_conversation_id UNIQUE (conversation_id, id),
                    CONSTRAINT uk_memori_conversation_message_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_conv_msg_conv
                       FOREIGN KEY (conversation_id)
                        REFERENCES memori_conversation (id)
                         ON DELETE CASCADE
                )
            """,
        },
        {
            "description": "create table memori_entity_fact",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_entity_fact(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    entity_id BIGINT NOT NULL,
                    content TEXT NOT NULL,
                    content_embedding BYTEA NOT NULL,
                    num_times BIGINT NOT NULL,
                    date_last_time TIMESTAMP NOT NULL,
                    uniq CHAR(64) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_entity_fact_entity_id UNIQUE (entity_id, id),
                    CONSTRAINT uk_memori_entity_fact_entity_id_uniq UNIQUE (entity_id, uniq),
                    CONSTRAINT uk_memori_entity_fact_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_ent_fact_entity
                       FOREIGN KEY (entity_id)
                        REFERENCES memori_entity (id)
                         ON DELETE CASCADE
                );
                CREATE INDEX IF NOT EXISTS idx_memori_entity_fact_entity_id_freq
                ON memori_entity_fact (entity_id, num_times DESC, date_last_time DESC);
                CREATE INDEX IF NOT EXISTS idx_memori_entity_fact_embedding_search
                ON memori_entity_fact (entity_id, id)
            """,
        },
        {
            "description": "create table memori_process_attribute",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_process_attribute(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    process_id BIGINT NOT NULL,
                    content TEXT NOT NULL,
                    num_times BIGINT NOT NULL,
                    date_last_time TIMESTAMP NOT NULL,
                    uniq CHAR(64) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_process_attribute_process_id UNIQUE (process_id, id),
                    CONSTRAINT uk_memori_process_attribute_process_id_uniq UNIQUE (process_id, uniq),
                    CONSTRAINT uk_memori_process_attribute_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_proc_attribute
                       FOREIGN KEY (process_id)
                        REFERENCES memori_process (id)
                         ON DELETE CASCADE
                )
            """,
        },
        {
            "description": "create table memori_subject",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_subject(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    name VARCHAR(255) NOT NULL,
                    type VARCHAR(255) NOT NULL,
                    uniq CHAR(64) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_subject_uniq UNIQUE (uniq),
                    CONSTRAINT uk_memori_subject_uuid UNIQUE (uuid)
                )
            """,
        },
        {
            "description": "create table memori_predicate",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_predicate(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    content TEXT NOT NULL,
                    uniq CHAR(64) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_predicate_uniq UNIQUE (uniq),
                    CONSTRAINT uk_memori_predicate_uuid UNIQUE (uuid)
                )
            """,
        },
        {
            "description": "create table memori_object",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_object(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    name VARCHAR(255) NOT NULL,
                    type VARCHAR(255) NOT NULL,
                    uniq CHAR(64) NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_object_uniq UNIQUE (uniq),
                    CONSTRAINT uk_memori_object_uuid UNIQUE (uuid)
                )
            """,
        },
        {
            "description": "create table memori_knowledge_graph",
            "operation": """
                CREATE TABLE IF NOT EXISTS memori_knowledge_graph(
                    id BIGSERIAL PRIMARY KEY,
                    uuid VARCHAR(36) NOT NULL,
                    entity_id BIGINT NOT NULL,
                    subject_id BIGINT NOT NULL,
                    predicate_id BIGINT NOT NULL,
                    object_id BIGINT NOT NULL,
                    num_times BIGINT NOT NULL,
                    date_last_time TIMESTAMP NOT NULL,
                    date_created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                    date_updated TIMESTAMP DEFAULT NULL,
                    --
                    CONSTRAINT uk_memori_knowledge_graph_entity_id UNIQUE (entity_id, id),
                    CONSTRAINT uk_memori_knowledge_graph_entity_subject_predicate_object UNIQUE (entity_id, subject_id, predicate_id, object_id),
                    CONSTRAINT uk_memori_knowledge_graph_object_id UNIQUE (object_id, id),
                    CONSTRAINT uk_memori_knowledge_graph_predicate_id UNIQUE (predicate_id, id),
                    CONSTRAINT uk_memori_knowledge_graph_subject_id UNIQUE (subject_id, id),
                    CONSTRAINT uk_memori_knowledge_graph_uuid UNIQUE (uuid),
                    --
                    CONSTRAINT fk_memori_know_graph_entity
                       FOREIGN KEY (entity_id)
                        REFERENCES memori_entity (id)
                         ON DELETE CASCADE,
                    CONSTRAINT fk_memori_know_graph_object
                       FOREIGN KEY (object_id)
                        REFERENCES memori_object (id)
                         ON DELETE CASCADE,
                    CONSTRAINT fk_memori_know_graph_predicate
                       FOREIGN KEY (predicate_id)
                        REFERENCES memori_predicate (id)
                         ON DELETE CASCADE,
                    CONSTRAINT fk_memori_know_graph_subject
                       FOREIGN KEY (subject_id)
                        REFERENCES memori_subject (id)
                         ON DELETE CASCADE
                )
            """,
        },
    ]
}



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import os
from collections.abc import Callable
from typing import Any
from uuid import uuid4

import psycopg

from memori._config import Config
from memori._exceptions import QuotaExceededError
from memori.llm._providers import Agno as LlmProviderAgno
from memori.llm._providers import Anthropic as LlmProviderAnthropic
from memori.llm._providers import Google as LlmProviderGoogle
from memori.llm._providers import LangChain as LlmProviderLangChain
from memori.llm._providers import OpenAi as LlmProviderOpenAi
from memori.llm._providers import PydanticAi as LlmProviderPydanticAi
from memori.llm._providers import XAi as LlmProviderXAi
from memori.memory.augmentation import Manager as AugmentationManager
from memori.memory.recall import Recall
from memori.storage import Manager as StorageManager

__all__ = ["Memori", "QuotaExceededError"]


class Memori:
    def __init__(self, conn: Callable[[], Any] | Any | None = None):
        self.config = Config()
        self.config.api_key = os.environ.get("MEMORI_API_KEY", None)
        self.config.enterprise = os.environ.get("MEMORI_ENTERPRISE", "0") == "1"
        self.config.session_id = uuid4()

        if conn is None:
            conn = self._get_default_connection()

        self.config.storage = StorageManager(self.config).start(conn)
        self.config.augmentation = AugmentationManager(self.config).start(conn)

        self.agno = LlmProviderAgno(self)
        self.anthropic = LlmProviderAnthropic(self)
        self.google = LlmProviderGoogle(self)
        self.langchain = LlmProviderLangChain(self)
        self.openai = LlmProviderOpenAi(self)
        self.pydantic_ai = LlmProviderPydanticAi(self)
        self.xai = LlmProviderXAi(self)

    def _get_default_connection(self) -> Callable[[], Any]:
        connection_string = os.environ.get("MEMORI_COCKROACHDB_CONNECTION_STRING")
        if connection_string:
            return lambda: psycopg.connect(connection_string)

        raise RuntimeError(
            "No connection factory provided. Either pass 'conn' parameter or set "
            "MEMORI_COCKROACHDB_CONNECTION_STRING environment variable."
        )

    def attribution(self, entity_id=None, process_id=None):
        if entity_id is not None:
            entity_id = str(entity_id)

            if len(entity_id) > 100:
                raise RuntimeError("entity_id cannot be greater than 100 characters")

        if process_id is not None:
            process_id = str(process_id)

            if len(process_id) > 100:
                raise RuntimeError("process_id cannot be greater than 100 characters")

        self.config.entity_id = entity_id
        self.config.process_id = process_id

        return self

    def new_session(self):
        self.config.session_id = uuid4()
        self.config.reset_cache()
        return self

    def set_session(self, id):
        self.config.session_id = id
        return self

    def recall(self, query: str, limit: int = 5):
        return Recall(self.config).search_facts(query, limit)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""


class Registry:
    _augmentations: dict[str, type] = {}

    @classmethod
    def register(cls, name: str):
        def decorator(augmentation_class: type):
            cls._augmentations[name] = augmentation_class
            return augmentation_class

        return decorator

    def augmentations(self, config=None):
        return [aug_class(config=config) for aug_class in self._augmentations.values()]



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from pymongo.synchronous.mongo_client import MongoClient

from memori.storage._base import BaseStorageAdapter
from memori.storage._registry import Registry


@Registry.register_adapter(
    lambda conn: hasattr(conn, "database") and hasattr(conn, "list_collection_names")
)
class Adapter(BaseStorageAdapter):
    """MongoDB storage adapter for MongoDB database connections."""

    def execute(self, collection_name_or_ops, operation=None, *args, **kwargs):
        """Execute MongoDB operations.

        Args:
            collection_name_or_ops: Collection name, list of ops, or single op dict
            operation: MongoDB operation (find_one, insert_one, etc.) - optional
            *args: Positional arguments for the operation
            **kwargs: Keyword arguments for the operation
        """
        if isinstance(self.conn, MongoClient):
            db = self.conn.get_default_database()
        else:
            db = self.conn

        if db is None:
            raise RuntimeError("MongoDB database connection is None")

        if operation is None:
            if isinstance(collection_name_or_ops, list):
                for op in collection_name_or_ops:
                    self._execute_operation(db, op)
            elif isinstance(collection_name_or_ops, dict):
                self._execute_operation(db, collection_name_or_ops)
            return None

        collection = db[collection_name_or_ops]
        return getattr(collection, operation)(*args, **kwargs)

    def commit(self):
        """MongoDB doesn't require explicit commits for single operations."""
        pass

    def flush(self):
        """MongoDB doesn't require explicit flushes for single operations."""
        pass

    def rollback(self):
        """MongoDB doesn't require explicit rollbacks for single operations."""
        pass

    def close(self):
        """MongoDB client connection should not be closed per-operation.

        The MongoClient is designed to be long-lived and shared across threads.
        Closing it would invalidate all connections from the client pool.
        """
        pass

    def get_dialect(self):
        return "mongodb"

    def _execute_operation(self, db, op):
        """Execute a single MongoDB operation from a dict.

        Args:
            db: MongoDB database instance
            op: Dict with 'collection', 'method', 'args', and 'kwargs' keys
        """
        collection = db[op["collection"]]
        method = getattr(collection, op["method"])
        args = op.get("args", [])
        kwargs = op.get("kwargs", {})
        method(*args, **kwargs)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import sys
from typing import Any

from memori._cli import Cli
from memori._config import Config
from memori._setup import Manager as SetupManager
from memori.api._quota import Manager as ApiQuotaManager
from memori.api._sign_up import Manager as ApiSignUpManager
from memori.storage.cockroachdb._cluster_manager import (
    ClusterManager as CockroachDBClusterManager,
)


def main():
    cli = Cli(Config())
    cli.banner()

    options: dict[str, dict[str, Any]] = {
        "cockroachdb": {
            "description": "Manager a CockroachDB cluster",
            "params": ["cluster", "<start | claim | delete>"],
            "obj": CockroachDBClusterManager,
        },
        "quota": {
            "description": "Check your quota",
            "params": [],
            "obj": ApiQuotaManager,
        },
        "setup": {
            "description": "Execute suggested setup steps",
            "params": [],
            "obj": SetupManager,
        },
        "sign-up": {
            "description": "Sign up for an API key",
            "params": ["<email_address>"],
            "obj": ApiSignUpManager,
        },
    }

    if len(sys.argv) <= 1 or sys.argv[1] not in options:
        cli.print("{:<15}{:<45}{:<6}".format("Option", "Description", "Params"))
        cli.print("{:<15}{:<45}{:<6}".format("------", "-----------", "------"))

        for key, value in options.items():
            params = value["params"]
            cli.print(
                "{:<15}{:<45}{:>6}".format(
                    key, value["description"], "Y" if len(params) > 0 else "N"
                )
            )

        cli.print("\nusage: python -m memori <option> [params]\n")
    else:
        option = options[sys.argv[1]]
        params = option["params"]
        obj_cls = option["obj"]
        if len(params) > 0:
            if len(sys.argv) != 2 + len(params):
                obj_cls(Config()).usage()
                cli.newline()
                sys.exit(1)

        obj_cls(Config()).execute()


if __name__ == "__main__":
    main()



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from datetime import datetime, timezone
from uuid import uuid4

from memori.storage._base import (
    BaseConversation,
    BaseConversationMessage,
    BaseConversationMessages,
    BaseEntity,
    BaseEntityFact,
    BaseKnowledgeGraph,
    BaseProcess,
    BaseProcessAttribute,
    BaseSchema,
    BaseSchemaVersion,
    BaseSession,
    BaseStorageAdapter,
)
from memori.storage._registry import Registry
from memori.storage.migrations._mongodb import migrations


class Conversation(BaseConversation):
    def __init__(self, conn: BaseStorageAdapter):
        super().__init__(conn)
        self.message = ConversationMessage(conn)
        self.messages = ConversationMessages(conn)

    def create(self, session_id, timeout_minutes: int):
        existing = self.conn.execute(
            "memori_conversation", "find_one", {"session_id": session_id}
        )

        if existing:
            last_message = self.conn.execute(
                "memori_conversation_message",
                "find_one",
                {"conversation_id": existing["_id"]},
                sort=[("date_created", -1)],
            )

            last_activity = (
                last_message["date_created"]
                if last_message
                else existing["date_created"]
            )

            now = datetime.now(timezone.utc)
            minutes_elapsed = (now - last_activity).total_seconds() / 60

            if minutes_elapsed <= timeout_minutes:
                return existing.get("_id")

        conversation_uuid = str(uuid4())
        conversation_doc = {
            "uuid": conversation_uuid,
            "session_id": session_id,
            "summary": None,
            "date_created": datetime.now(timezone.utc),
            "date_updated": None,
        }

        result = self.conn.execute(
            "memori_conversation", "insert_one", conversation_doc
        )

        return result.inserted_id

    def update(self, id: int, summary: str):
        if summary is None:
            return self

        self.conn.execute(
            "memori_conversation",
            "update_one",
            {"_id": id},
            {"$set": {"summary": summary}},
        )

        return self

    def read(self, id: int) -> dict | None:
        result = self.conn.execute(
            "memori_conversation",
            "find_one",
            {"_id": id},
        )

        if result is None:
            return None

        # Convert MongoDB result to dict, excluding _id or converting it to id
        conversation = dict(result)
        if "_id" in conversation:
            conversation["id"] = conversation.pop("_id")

        return conversation


class ConversationMessage(BaseConversationMessage):
    def create(self, conversation_id: int, role: str, type: str, content: str):
        message_doc = {
            "uuid": str(uuid4()),
            "conversation_id": conversation_id,
            "role": role,
            "type": type,
            "content": content,
            "date_created": datetime.now(timezone.utc),
            "date_updated": None,
        }

        self.conn.execute("memori_conversation_message", "insert_one", message_doc)


class ConversationMessages(BaseConversationMessages):
    def read(self, conversation_id: int):
        results = self.conn.execute(
            "memori_conversation_message",
            "find",
            {"conversation_id": conversation_id},
            {"role": 1, "content": 1, "_id": 0},
        )

        messages = []
        for result in results:
            messages.append({"content": result["content"], "role": result["role"]})

        return messages


class Entity(BaseEntity):
    def create(self, external_id: str):
        # Check if entity already exists
        existing = self.conn.execute(
            "memori_entity", "find_one", {"external_id": external_id}
        )

        if existing:
            return existing.get("_id")

        # Create new entity
        entity_doc = {
            "uuid": str(uuid4()),
            "external_id": external_id,
            "date_created": datetime.now(timezone.utc),
            "date_updated": None,
        }

        result = self.conn.execute("memori_entity", "insert_one", entity_doc)

        return result.inserted_id


class EntityFact(BaseEntityFact):
    def create(self, entity_id: int, facts: list, fact_embeddings: list | None = None):
        if facts is None or len(facts) == 0:
            return self

        from memori._utils import generate_uniq
        from memori.llm._embeddings import format_embedding_for_db

        for i, fact in enumerate(facts):
            embedding = (
                fact_embeddings[i]
                if fact_embeddings and i < len(fact_embeddings)
                else []
            )
            embedding_formatted = format_embedding_for_db(embedding, "mongodb")
            uniq = generate_uniq([fact])

            # Check if fact already exists
            existing = self.conn.execute(
                "memori_entity_fact",
                "find_one",
                {"entity_id": entity_id, "uniq": uniq},
            )

            if existing:
                # Update existing fact
                self.conn.execute(
                    "memori_entity_fact",
                    "update_one",
                    {"_id": existing["_id"]},
                    {
                        "$inc": {"num_times": 1},
                        "$set": {"date_last_time": datetime.now(timezone.utc)},
                    },
                )
            else:
                # Insert new fact
                fact_doc = {
                    "uuid": str(uuid4()),
                    "entity_id": entity_id,
                    "content": fact,
                    "content_embedding": embedding_formatted,
                    "num_times": 1,
                    "date_last_time": datetime.now(timezone.utc),
                    "uniq": uniq,
                    "date_created": datetime.now(timezone.utc),
                    "date_updated": None,
                }

                self.conn.execute("memori_entity_fact", "insert_one", fact_doc)

        return self

    def get_embeddings(self, entity_id: int, limit: int = 1000):
        results = self.conn.execute(
            "memori_entity_fact",
            "find",
            {"entity_id": entity_id},
            {"_id": 1, "content_embedding": 1},
        )

        embeddings = []
        for result in list(results)[:limit]:
            embeddings.append(
                {"id": result["_id"], "content_embedding": result["content_embedding"]}
            )

        return embeddings

    def get_facts_by_ids(self, fact_ids: list[int]):
        if not fact_ids:
            return []

        results = self.conn.execute(
            "memori_entity_fact",
            "find",
            {"_id": {"$in": fact_ids}},
            {"_id": 1, "content": 1},
        )

        facts = []
        for result in results:
            facts.append({"id": result["_id"], "content": result["content"]})

        return facts


class KnowledgeGraph(BaseKnowledgeGraph):
    def create(self, entity_id: int, semantic_triples: list):
        if semantic_triples is None or len(semantic_triples) == 0:
            return self

        from datetime import datetime, timezone

        from memori._utils import generate_uniq

        for semantic_triple in semantic_triples:
            # Insert or get subject
            subject_uniq = generate_uniq(
                [semantic_triple.subject_name, semantic_triple.subject_type]
            )
            existing_subject = self.conn.execute(
                "memori_subject", "find_one", {"uniq": subject_uniq}
            )

            if existing_subject:
                subject_id = existing_subject["_id"]
            else:
                subject_doc = {
                    "uuid": str(uuid4()),
                    "name": semantic_triple.subject_name,
                    "type": semantic_triple.subject_type,
                    "uniq": subject_uniq,
                    "date_created": datetime.now(timezone.utc),
                    "date_updated": None,
                }
                result = self.conn.execute("memori_subject", "insert_one", subject_doc)
                subject_id = result.inserted_id

            # Insert or get predicate
            predicate_uniq = generate_uniq([semantic_triple.predicate])
            existing_predicate = self.conn.execute(
                "memori_predicate", "find_one", {"uniq": predicate_uniq}
            )

            if existing_predicate:
                predicate_id = existing_predicate["_id"]
            else:
                predicate_doc = {
                    "uuid": str(uuid4()),
                    "content": semantic_triple.predicate,
                    "uniq": predicate_uniq,
                    "date_created": datetime.now(timezone.utc),
                    "date_updated": None,
                }
                result = self.conn.execute(
                    "memori_predicate", "insert_one", predicate_doc
                )
                predicate_id = result.inserted_id

            # Insert or get object
            object_uniq = generate_uniq(
                [semantic_triple.object_name, semantic_triple.object_type]
            )
            existing_object = self.conn.execute(
                "memori_object", "find_one", {"uniq": object_uniq}
            )

            if existing_object:
                object_id = existing_object["_id"]
            else:
                object_doc = {
                    "uuid": str(uuid4()),
                    "name": semantic_triple.object_name,
                    "type": semantic_triple.object_type,
                    "uniq": object_uniq,
                    "date_created": datetime.now(timezone.utc),
                    "date_updated": None,
                }
                result = self.conn.execute("memori_object", "insert_one", object_doc)
                object_id = result.inserted_id

            # Insert or update knowledge graph entry
            if (
                entity_id is not None
                and subject_id is not None
                and predicate_id is not None
                and object_id is not None
            ):
                existing_kg = self.conn.execute(
                    "memori_knowledge_graph",
                    "find_one",
                    {
                        "entity_id": entity_id,
                        "subject_id": subject_id,
                        "predicate_id": predicate_id,
                        "object_id": object_id,
                    },
                )

                if existing_kg:
                    self.conn.execute(
                        "memori_knowledge_graph",
                        "update_one",
                        {"_id": existing_kg["_id"]},
                        {
                            "$inc": {"num_times": 1},
                            "$set": {"date_last_time": datetime.now(timezone.utc)},
                        },
                    )
                else:
                    kg_doc = {
                        "uuid": str(uuid4()),
                        "entity_id": entity_id,
                        "subject_id": subject_id,
                        "predicate_id": predicate_id,
                        "object_id": object_id,
                        "num_times": 1,
                        "date_last_time": datetime.now(timezone.utc),
                        "date_created": datetime.now(timezone.utc),
                        "date_updated": None,
                    }
                    self.conn.execute("memori_knowledge_graph", "insert_one", kg_doc)

        return self


class Process(BaseProcess):
    def create(self, external_id: str):
        # Check if process already exists
        existing = self.conn.execute(
            "memori_process", "find_one", {"external_id": external_id}
        )

        if existing:
            return existing.get("_id")

        # Create new process
        process_doc = {
            "uuid": str(uuid4()),
            "external_id": external_id,
            "date_created": datetime.now(timezone.utc),
            "date_updated": None,
        }

        result = self.conn.execute("memori_process", "insert_one", process_doc)

        return result.inserted_id


class ProcessAttribute(BaseProcessAttribute):
    def create(self, process_id: int, attributes: list):
        if attributes is None or len(attributes) == 0:
            return self

        from datetime import datetime, timezone

        from memori._utils import generate_uniq

        for attribute in attributes:
            uniq = generate_uniq([attribute])
            existing = self.conn.execute(
                "memori_process_attribute",
                "find_one",
                {"process_id": process_id, "uniq": uniq},
            )

            if existing:
                self.conn.execute(
                    "memori_process_attribute",
                    "update_one",
                    {"_id": existing["_id"]},
                    {
                        "$inc": {"num_times": 1},
                        "$set": {"date_last_time": datetime.now(timezone.utc)},
                    },
                )
            else:
                attribute_doc = {
                    "uuid": str(uuid4()),
                    "process_id": process_id,
                    "content": attribute,
                    "num_times": 1,
                    "date_last_time": datetime.now(timezone.utc),
                    "uniq": uniq,
                    "date_created": datetime.now(timezone.utc),
                    "date_updated": None,
                }
                self.conn.execute(
                    "memori_process_attribute", "insert_one", attribute_doc
                )

        return self


class Session(BaseSession):
    def create(self, uuid: str, entity_id: int, process_id: int):
        # Check if session already exists
        existing = self.conn.execute("memori_session", "find_one", {"uuid": str(uuid)})

        if existing:
            return existing.get("_id")

        # Create new session
        session_doc = {
            "uuid": str(uuid),
            "entity_id": entity_id,
            "process_id": process_id,
            "date_created": datetime.now(timezone.utc),
            "date_updated": None,
        }

        result = self.conn.execute("memori_session", "insert_one", session_doc)

        return result.inserted_id


class Schema(BaseSchema):
    def __init__(self, conn: BaseStorageAdapter):
        super().__init__(conn)
        self.version = SchemaVersion(conn)


class SchemaVersion(BaseSchemaVersion):
    def create(self, num: int):
        schema_doc = {"num": num}

        self.conn.execute("memori_schema_version", "insert_one", schema_doc)

    def delete(self):
        self.conn.execute("memori_schema_version", "delete_many", {})

    def read(self):
        result = self.conn.execute(
            "memori_schema_version", "find_one", {}, {"num": 1, "_id": 0}
        )

        if not result:
            return None

        return result.get("num")


@Registry.register_driver("mongodb")
class Driver:
    """MongoDB storage driver.

    Attributes:
        migrations: Database schema migrations for MongoDB.
        requires_rollback_on_error: MongoDB does not abort transactions on query
            errors by default, so no rollback is needed to continue executing queries.
    """

    migrations = migrations
    requires_rollback_on_error = False

    def __init__(self, conn: BaseStorageAdapter):
        self.conversation = Conversation(conn)
        self.entity = Entity(conn)
        self.entity_fact = EntityFact(conn)
        self.knowledge_graph = KnowledgeGraph(conn)
        self.process = Process(conn)
        self.process_attribute = ProcessAttribute(conn)
        self.schema = Schema(conn)
        self.session = Session(conn)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

import time

from memori.llm._base import BaseIterator
from memori.memory._manager import Manager as MemoryManager


class AsyncIterator(BaseIterator):
    def __aiter__(self):
        self.iterator = self.source_iterator.__aiter__()
        return self

    async def __anext__(self):
        try:
            if self.iterator is None:
                raise RuntimeError("Iterator not initialized")
            chunk = await self.iterator.__anext__()

            self.set_raw_response()
            self.process_chunk(chunk)

            return chunk
        except StopAsyncIteration:
            MemoryManager(self.config).execute(
                self.invoke._format_payload(
                    self.config.framework.provider,
                    self.config.llm.provider,
                    self.config.llm.version,
                    self._time_start,
                    time.time(),
                    self.invoke._format_kwargs(self._kwargs),
                    self.invoke._format_response(self.raw_response),
                )
            )
            raise

    async def __aenter__(self):
        await self.source_iterator.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        return await self.source_iterator.__aexit__(exc_type, exc, tb)


class Iterator(BaseIterator):
    def __iter__(self):
        return self

    def __next__(self):
        try:
            chunk = next(self.source_iterator)

            self.set_raw_response()
            self.process_chunk(chunk)

            return chunk
        except StopIteration:
            MemoryManager(self.config).execute(
                self.invoke._format_payload(
                    self.config.framework.provider,
                    self.config.llm.provider,
                    self.config.llm.version,
                    self._time_start,
                    time.time(),
                    self.invoke._format_kwargs(self._kwargs),
                    self.invoke._format_response(self.raw_response),
                )
            )

            raise

    def __enter__(self):
        self.source_iterator.__enter__()
        return self

    def __exit__(self, exc_type, exc, tb):
        return self.source_iterator.__exit__(exc_type, exc, tb)



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                 perfectam memoriam
                      memorilabs.ai
"""

import time

from sqlalchemy.exc import OperationalError

from memori._config import Config
from memori._search import search_entity_facts
from memori.llm._embeddings import embed_texts

MAX_RETRIES = 3
RETRY_BACKOFF_BASE = 0.05


class Recall:
    def __init__(self, config: Config) -> None:
        self.config = config

    def search_facts(
        self, query: str, limit: int | None = None, entity_id: int | None = None
    ) -> list[dict]:
        if self.config.storage is None or self.config.storage.driver is None:
            return []

        if entity_id is None:
            if self.config.entity_id is None:
                return []
            entity_id = self.config.storage.driver.entity.create(self.config.entity_id)

        if entity_id is None:
            return []

        if limit is None:
            limit = self.config.recall_facts_limit

        query_embedding = embed_texts(query)[0]

        facts = []
        for attempt in range(MAX_RETRIES):
            try:
                facts = search_entity_facts(
                    self.config.storage.driver.entity_fact,
                    entity_id,
                    query_embedding,
                    limit,
                    self.config.recall_embeddings_limit,
                )
                break
            except OperationalError as e:
                if "restart transaction" in str(e) and attempt < MAX_RETRIES - 1:
                    time.sleep(RETRY_BACKOFF_BASE * (2**attempt))
                    continue
                raise

        return facts



r"""
 __  __                           _
|  \/  | ___ _ __ ___   ___  _ __(_)
| |\/| |/ _ \ '_ ` _ \ / _ \| '__| |
| |  | |  __/ | | | | | (_) | |  | |
|_|  |_|\___|_| |_| |_|\___/|_|  |_|
                  perfectam memoriam
                       memorilabs.ai
"""

from memori.storage._manager import Manager

# Import adapters and drivers to trigger their self-registration decorators.
# Order matters: more specific matchers (sqlalchemy, django) before generic ones (mongodb, dbapi)
from memori.storage.adapters import sqlalchemy, django, mongodb, dbapi  # noqa: F401
from memori.storage.drivers import (
    mongodb as mongodb_driver,  # noqa: F401
)
from memori.storage.drivers import (
    mysql,  # noqa: F401
    oracle,  # noqa: F401
    postgresql,  # noqa: F401
    sqlite,  # noqa: F401
)

__all__ = ["Manager"]
